from astropy import time

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-26T12:06:21.930', format='isot', scale='utc')

start_time = time.Time(1240327333.348145, format='gps', scale='utc')
end_time = time.Time('2019-04-27T04:21:39.991', format='isot', scale='utc')

start_time = time.Time('2019-04-27T04:21:39.991', format='isot', scale='utc')
end_time = time.Time('2019-04-28T11:41:07.848', format='isot', scale='utc')

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-25T09:39:48', format='isot', scale='utc')

start_time = time.Time('2019-05-10T02:59:39.292', format='isot', scale='utc')
end_time = time.Time('2019-05-10T06:00:25', format='isot', scale='utc')

start_time = time.Time('2019-05-10T06:00:25', format='isot', scale='utc')
end_time = time.Time('2019-05-11T00:31:40', format='isot', scale='utc')

start_time = time.Time('2019-04-26T15:21:55', format='isot', scale='utc')
end_time = time.Time('2019-04-26T17:23:04.047', format='isot', scale='utc')

start_time = time.Time('2019-04-26T17:23:04.047', format='isot', scale='utc')
end_time = time.Time('2019-04-27T22:47:04.528', format='isot', scale='utc')

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-25T09:06:42.775', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-04-25T09:06:42.775', format='isot', scale='utc') 
end_time = time.Time('2019-04-25T12:51:34', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-04-25T12:51:34', format='isot', scale='utc')
end_time = time.Time('2019-04-25T14:51:57', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T21:32:46', format='isot', scale='utc')
end_time = time.Time('2019-08-15T04:15:06', format='isot', scale='utc')

start_time = time.Time('2019-08-14T21:10:39.013', format='isot', scale='utc')
end_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T04:19:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T00:09:41', format='isot', scale='utc')
end_time = time.Time('2019-08-15T02:38:20', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T05:39:26', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T10:32:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T10:32:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T12:11:49.075', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time(58710.22978, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time(58710.22978, format='mjd', scale='utc')
end_time = time.Time(58710.44370, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time(58710.5, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time(58710.5, format='mjd', scale='utc')
end_time = time.Time(58711.5, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T08:07:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T12:40:37', format='isot', scale='utc')
end_time = time.Time('2019-08-15T15:13:38', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-01T23:31:01.838', format='isot', scale='utc')
end_time = time.Time('2019-09-01T23:37:46', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-01T23:37:46', format='isot', scale='utc')
end_time = time.Time('2019-09-04T05:45:10', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-10T01:26:19.243', format='isot', scale='utc')
end_time = time.Time('2019-09-10T02:58:42.5', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)

start_time = time.Time('2019-09-10T08:29:58.544', format='isot', scale='utc')
end_time = time.Time('2019-09-10T10:18:00', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)
